import json
import logging

from werkzeug.wrappers import Request, Response
from werkzeug.exceptions import HTTPException, abort

import alexandra.util as util


log = logging.getLogger(__name__)


class WsgiApp:
    """This class is wraps :py:obj:`alexandra.app.Application` object and
    handles all the gory details of parsing Alexa requests and validating that
    they were actually sent by Amazon.

    This class implements the WSGI interface, so an instance of can be passed
    to standard WSGI servers (uWSGI, gunicorn, ...)
    """

    def __init__(self, alexa, validate_requests=True):
        """
        :param alexa: alexandra.app.App to wrap
        :param validate_requests: Whether or not to do timestamp and
            certificate validation.
        """

        self.alexa = alexa
        self.validate = validate_requests

    # The Request.application decorator handles some boring parts of making
    # this work with WSGI
    @Request.application
    def __call__(self, request):
        return self.wsgi_app(request)

    def wsgi_app(self, request):
        """Incoming request handler.

        :param request: Werkzeug request object
        """

        try:
            if request.method != 'POST':
                abort(400)

            try:
                # Python 2.7 compatibility
                data = request.data
                if isinstance(data, str):
                    body = json.loads(data)
                else:
                    body = json.loads(data.decode('utf-8'))
            except ValueError:
                abort(400)

            if self.validate:
                valid_cert = util.validate_request_certificate(
                    request.headers, request.data)

                valid_ts = util.validate_request_timestamp(body)

                if not valid_cert or not valid_ts:
                    log.error('failed to validate request')
                    abort(403)

            resp_obj = self.alexa.dispatch_request(body)
            return Response(response=json.dumps(resp_obj, indent=4),
                            status=200,
                            mimetype='application/json')

        except HTTPException as exc:
            log.exception('Failed to handle request')
            return exc



# flake8: noqa
"""
Python support for Alexa applications.

Because like everything Amazon it involves a ton of tedious boilerplate.
"""

import logging
logging.getLogger(__name__).addHandler(logging.NullHandler())


from alexandra.app import Application
from alexandra.session import Session
from alexandra.util import respond, reprompt



import logging

from werkzeug.serving import run_simple

from alexandra.session import Session
from alexandra.util import respond
from alexandra.wsgi import WsgiApp

log = logging.getLogger(__name__)


class Application:
    def __init__(self):
        self.intent_map = {}
        self.unknown_intent_fn = lambda x, y: respond(text='unknown intent')
        self.launch_fn = lambda _: respond()
        self.session_end_fn = respond

    def create_wsgi_app(self, validate_requests=True):
        """Return an object that can be run by any WSGI server (uWSGI,
        etc.) to serve this Alexa application.
        """

        return WsgiApp(self, validate_requests)

    def run(self, host, port, debug=True, validate_requests=True):
        """Utility method to quickly get a server up and running.

        :param debug: turns on Werkzeug debugger, code reloading, and full
            logging.
        :param validate_requests: whether or not to ensure that requests are
            sent by Amazon. This can be usefulfor manually testing the server.
        """

        if debug:
            # Turn on all alexandra log output
            logging.basicConfig(level=logging.DEBUG)

        app = self.create_wsgi_app(validate_requests)
        run_simple(host, port, app, use_reloader=debug, use_debugger=debug)

    def dispatch_request(self, body):
        """Given a parsed JSON request object, call the correct Intent, Launch,
        or SessionEnded function.

        This function is called after request parsing and validaion and will
        raise a `ValueError` if an unknown request type comes in.

        :param body: JSON object loaded from incoming request's POST data.
        """

        req_type = body.get('request', {}).get('type')
        session_obj = body.get('session')

        session = Session(session_obj) if session_obj else None

        if req_type == 'LaunchRequest':
            return self.launch_fn(session)

        elif req_type == 'IntentRequest':
            intent = body['request']['intent']['name']
            intent_fn = self.intent_map.get(intent, self.unknown_intent_fn)

            slots = {
                slot['name']: slot.get('value')
                for _, slot in
                body['request']['intent'].get('slots', {}).items()
            }

            arity = intent_fn.__code__.co_argcount

            if arity == 2:
                return intent_fn(slots, session)

            return intent_fn()

        elif req_type == 'SessionEndedRequest':
            return self.session_end_fn()

        log.error('invalid request type: %s', req_type)
        raise ValueError('bad request: %s', body)

    def launch(self, func):
        """Decorator to register a function to be called whenever the
        app receives a LaunchRequest (which happens when someone
        invokes your skill without specifying an intent). ::

            @alexa_app.launch
            def launch_handler(session):
                pass
        """

        self.launch_fn = func
        return func

    def intent(self, intent_name):
        """Decorator to register a handler for the given intent.

        The decorated function can either take 0 or 2 arguments. If two are
        specified, it will be provided a dictionary of `{slot_name: value}` and
        a :py:class:`alexandra.session.Session` instance.

        If no session was provided in the request, the session object will be
        `None`. ::

            @alexa_app.intent('FooBarBaz')
            def foo_bar_baz_intent(slots, session):
                pass

            @alexa_app.intent('NoArgs')
            def noargs_intent():
                pass
        """

        # nested decorator so we can have params.
        def _decorator(func):
            arity = func.__code__.co_argcount

            if arity not in [0, 2]:
                raise ValueError("expected 0 or 2 argument function")

            self.intent_map[intent_name] = func
            return func

        return _decorator

    def unknown_intent(self, func):
        """Decorator to register a function to be called when an unknown intent
        is received. This should only happen when the intents/utterance file
        are malformed.
        """

        self.unknown_intent_fn = func
        return func

    def session_end(self, func):
        """ Decorator to register a function to be called when a
        SessionEndedRequest is received.
        """

        self.session_end_fn = func
        return func



"""Utility functionality for Alexandra"""

import base64
import logging
import posixpath

try:
    from urlparse import urlparse
    from urllib2 import urlopen
except:
    from urllib.parse import urlparse
    from urllib.request import urlopen

from datetime import datetime

from OpenSSL import crypto


# We don't want to check the certificate every single time. Store for
# as long as they are valid.
_cache = {}
log = logging.getLogger(__name__)


def respond(text=None, ssml=None, attributes=None, reprompt_text=None,
            reprompt_ssml=None, end_session=True):
    """ Build a dict containing a valid response to an Alexa request.

    If speech output is desired, either of `text` or `ssml` should
    be specified.

    :param text: Plain text speech output to be said by Alexa device.
    :param ssml: Speech output in SSML form.
    :param attributes: Dictionary of attributes to store in the session.
    :param end_session: Should the session be terminated after this response?
    :param reprompt_text, reprompt_ssml: Works the same as
        `text`/`ssml`, but instead sets the reprompting speech output.
    """

    obj = {
        'version': '1.0',
        'response': {
            'outputSpeech': {'type': 'PlainText', 'text': ''},
            'shouldEndSession': end_session
        },
        'sessionAttributes': attributes or {}
    }

    if text:
        obj['response']['outputSpeech'] = {'type': 'PlainText', 'text': text}
    elif ssml:
        obj['response']['outputSpeech'] = {'type': 'SSML', 'ssml': ssml}

    reprompt_output = None
    if reprompt_text:
        reprompt_output = {'type': 'PlainText', 'text': reprompt_text}
    elif reprompt_ssml:
        reprompt_output = {'type': 'SSML', 'ssml': reprompt_ssml}

    if reprompt_output:
        obj['response']['reprompt'] = {'outputSpeech': reprompt_output}

    return obj


def reprompt(text=None, ssml=None, attributes=None):
    """Convenience method to save a little bit of typing for the common case of
    reprompting the user. Simply calls :py:func:`alexandra.util.respond` with
    the given arguments and holds the session open.

    One of either the `text` or `ssml` should be provided if any
    speech output is desired.

    :param text: Plain text speech output
    :param ssml: Speech output in SSML format
    :param attributes: Dictionary of attributes to store in the current session
    """

    return respond(
        reprompt_text=text,
        reprompt_ssml=ssml,
        attributes=attributes,
        end_session=False
    )


def validate_request_timestamp(req_body, max_diff=150):
    """Ensure the request's timestamp doesn't fall outside of the
    app's specified tolerance.

    Returns True if this request is valid, False otherwise.

    :param req_body: JSON object parsed out of the raw POST data of a request.
    :param max_diff: Maximum allowable difference in seconds between request
        timestamp and system clock. Amazon requires <= 150 seconds for
        published skills.
    """

    time_str = req_body.get('request', {}).get('timestamp')

    if not time_str:
        log.error('timestamp not present %s', req_body)
        return False

    req_ts = datetime.strptime(time_str, "%Y-%m-%dT%H:%M:%SZ")
    diff = (datetime.utcnow() - req_ts).total_seconds()

    if abs(diff) > max_diff:
        log.error('timestamp difference too high: %d sec', diff)
        return False

    return True


def validate_request_certificate(headers, data):
    """Ensure that the certificate and signature specified in the
    request headers are truely from Amazon and correctly verify.

    Returns True if certificate verification succeeds, False otherwise.

    :param headers: Dictionary (or sufficiently dictionary-like) map of request
        headers.
    :param data: Raw POST data attached to this request.
    """

    # Make sure we have the appropriate headers.
    if 'SignatureCertChainUrl' not in headers or \
       'Signature' not in headers:
        log.error('invalid request headers')
        return False

    cert_url = headers['SignatureCertChainUrl']
    sig = base64.b64decode(headers['Signature'])

    cert = _get_certificate(cert_url)

    if not cert:
        return False

    try:
        # ... wtf kind of API decision is this
        crypto.verify(cert, sig, data, 'sha1')
        return True
    except:
        log.error('invalid request signature')
        return False


def _get_certificate(cert_url):
    """Download and validate a specified Amazon PEM file."""
    global _cache

    if cert_url in _cache:
        cert = _cache[cert_url]
        if cert.has_expired():
            _cache = {}
        else:
            return cert

    url = urlparse(cert_url)
    host = url.netloc.lower()
    path = posixpath.normpath(url.path)

    # Sanity check location so we don't get some random person's cert.
    if url.scheme != 'https' or \
       host not in ['s3.amazonaws.com', 's3.amazonaws.com:443'] or \
       not path.startswith('/echo.api/'):
        log.error('invalid cert location %s', cert_url)
        return

    resp = urlopen(cert_url)
    if resp.getcode() != 200:
        log.error('failed to download certificate')
        return

    cert = crypto.load_certificate(crypto.FILETYPE_PEM, resp.read())

    if cert.has_expired() or cert.get_subject().CN != 'echo-api.amazon.com':
        log.error('certificate expired or invalid')
        return

    _cache[cert_url] = cert
    return cert



from setuptools import setup

setup(
    name='alexandra',
    version='0.4.0',
    description='Toolkit for writing Amazon Alexa skills as web services',
    author='Erik Price',
    url='https://github.com/erik/alexandra',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: ISC License (ISCL)',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.6',
        'Topic :: Home Automation',
        'Topic :: Internet :: WWW/HTTP :: WSGI :: Application',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
    ],
    packages=['alexandra'],
    license='ISC',
    test_requires=['tox'],
    install_requires=[
        'Werkzeug==0.11.4',
        'pyOpenssl==17.5.0'
    ]
)



import json


class Session:
    """Provides easier access to session objects sent along with requests.

    The object parsed from the request can be accessed directly
    through the :py:attr:`body` member.
    """

    def __init__(self, session_body):
        self.body = session_body

    def __repr__(self):
        return '<Session %s>' % json.dumps(self.body)

    @property
    def is_new(self):
        """Is this a new session or a previously open one?"""
        return self.body['new']

    @property
    def session_id(self):
        return self.body['sessionId']

    @property
    def application_id(self):
        return self.body['application']['applicationId']

    @property
    def user_id(self):
        return self.body['user'].get('userId')

    @property
    def user_access_token(self):
        return self.body['user'].get('accessToken')

    def get(self, attr, default=None):
        """Get an attribute defined by this session"""

        attrs = self.body.get('attributes') or {}
        return attrs.get(attr, default)
